package com.bureau.behavioralbiometrics

import android.app.Application
import android.os.Build
import android.os.SystemClock
import android.util.Log
import androidx.annotation.Keep
import com.bureau.base.Environment
import com.bureau.behavioralbiometrics.applifecycleevents.ForegroundStateManager
import com.bureau.behavioralbiometrics.applifecycleevents.models.AppLifecycleEventLocal
import com.bureau.behavioralbiometrics.applifecycleevents.models.toAppLifeCyEvents
import com.bureau.behavioralbiometrics.data.BBRepository
import com.bureau.behavioralbiometrics.data.remote.protoModels.bundle
import com.bureau.behavioralbiometrics.di.BBComponent
import com.bureau.behavioralbiometrics.di.COROUTINE_SCOPE_IO
import com.bureau.behavioralbiometrics.di.COROUTINE_SCOPE_MAIN
import com.bureau.behavioralbiometrics.di.DaggerBBComponent
import com.bureau.behavioralbiometrics.focuschangedata.models.toFocuChangeData
import com.bureau.behavioralbiometrics.frameratecollection.models.toFrameRateEvent
import com.bureau.behavioralbiometrics.keypressTypedata.models.toInputChangeData
import com.bureau.behavioralbiometrics.keypressTypedata.models.toKeyPressData
import com.bureau.behavioralbiometrics.models.BundleLocal
import com.bureau.behavioralbiometrics.sensordata.models.toAccelerometerData
import com.bureau.behavioralbiometrics.sensordata.models.toGyroscopeData
import com.bureau.behavioralbiometrics.touchtypedata.models.toPointerData
import com.bureau.behavioralbiometrics.touchtypedata.models.toViewportScrollData
import com.bureau.behavioralbiometrics.utils.TimeUtils
import com.bureau.behavioralbiometrics.websocket.WebSocket
import io.sentry.Sentry
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Job
import kotlinx.coroutines.delay
import kotlinx.coroutines.flow.first
import kotlinx.coroutines.isActive
import kotlinx.coroutines.launch
import org.json.JSONObject
import java.util.Base64
import java.util.UUID
import javax.inject.Inject
import javax.inject.Named

@Keep
class UserBehaviorCapture private constructor(
    application: Application,
    environment: Environment,
    credId: String,
    private val sessionId: String,
    private val sendExceptionToSentry: Boolean,
    isDev: Boolean = false
) {
    @Inject
    internal lateinit var bbRepository: BBRepository

    @Inject
    @Named(COROUTINE_SCOPE_IO)
    internal lateinit var coroutineScopeIO: CoroutineScope

    @Inject
    @Named(COROUTINE_SCOPE_MAIN)
    internal lateinit var coroutineScopeMain: CoroutineScope

    @Inject
    internal lateinit var foregroundStateManager: ForegroundStateManager

    @Inject
    internal lateinit var webSocket: WebSocket

    internal var bbComponent: BBComponent

    private var job1: Job? = null
    private var job2: Job? = null
    private var job3: Job? = null
    private var job4: Job? = null
    private var job5: Job? = null
    private var job6: Job? = null

    init {
        INSTANCE = this
        bbComponent =
            DaggerBBComponent.factory().create(application, environment, credId, sessionId, isDev).apply {
                inject(this@UserBehaviorCapture)
            }
        webSocket.webSocketService
    }

    @Keep
    fun startSubSession(subSessionId: String, userId: String) {
        stopSubSession(skipDelay = true) {
            val windowCallback = foregroundStateManager.windowCallback
            windowCallback?.isSessionActive = true
            if (windowCallback != null && !windowCallback.isCapturing) {
                windowCallback.captureData()
            }

            job1 = coroutineScopeIO.launch {
                foregroundStateManager.foregroundStateFlow.collect {
                    bbRepository.insertAppLifecycleEvent(
                        AppLifecycleEventLocal(
                            timeStamp = SystemClock.elapsedRealtime(),
                            appLifecycleStatus = it
                        )
                    )
                }
            }

            job2 = coroutineScopeIO.launch {
                foregroundStateManager.accelerometerDataFlow.collect {
                    bbRepository.insertAccelerometerData(it)
                }
            }

            job3 = coroutineScopeIO.launch {
                foregroundStateManager.gyroscopeDataFlow.collect {
                    bbRepository.insertGyroscopeData(it)
                }
            }

            job4 = coroutineScopeIO.launch {
                foregroundStateManager.fpsFrameCallback.frameRateEventLocalFlow.collect {
                    bbRepository.insertFrameRateEvent(it)
                }
            }

            var bundleIndex = 0L
            bundleMap.clear()

            job5 = coroutineScopeIO.launch {
                if (bbRepository.registerSubSession(sessionId, subSessionId, userId)) {
                    webSocket.webSocketStatus.collect { webSocketStatus ->
                        while (webSocketStatus == WebSocket.WebSocketStatus.OPEN && isActive) {
                            val appLifeCycleEvents =
                                bbRepository.getAllAppLifecycleEventsFlow().first()
                            val touchTypeData = bbRepository.getAllTouchTypeDataFlow().first()
                            val keyPressEvents = bbRepository.getAllKeyPressEventFlow().first()
                            val inputChangeEvents =
                                bbRepository.getAllInputChangeEventFlow().first()
                            val focusChangeData = bbRepository.getAllFocusChangeDataFlow().first()
                            val viewPortScrollData =
                                bbRepository.getAllViewPortScrollDataFlow().first()
                            val frameRateEvents = bbRepository.getAllFrameRateEventFlow().first()
                            val accelerometerDataLocal = bbRepository.getAllAccelerometerDataFlow().first()
                            val gyroscopeDataLocal = bbRepository.getAllGyroscopeDataFlow().first()
                            val id = UUID.randomUUID().toString().replace("-", "").substring(0, 8)

                            val bundle = bundle {
                                sessionId = this@UserBehaviorCapture.sessionId
                                clientTime = TimeUtils.getTimeSinceLastRebootMillis()
                                this.subSessionId = subSessionId
                                index = bundleIndex
                                this.id = id
                                appLifeCycleEvents.forEach {
                                    this.appLifeCycleEvents.add(it.toAppLifeCyEvents())
                                }
                                touchTypeData.forEach {
                                    this.pointerData.add(it.toPointerData())
                                }
                                keyPressEvents.forEach {
                                    this.keyPressData.add(it.toKeyPressData())
                                }
                                inputChangeEvents.forEach {
                                    this.inputChangeData.add(it.toInputChangeData())
                                }
                                focusChangeData.forEach {
                                    this.focusChangeData.add(it.toFocuChangeData())
                                }
                                viewPortScrollData.forEach {
                                    this.viewportScrollData.add(it.toViewportScrollData())
                                }
                                frameRateEvents.forEach {
                                    this.frameRateEvent.add(it.toFrameRateEvent())
                                }
                                accelerometerDataLocal.forEach {
                                    this.accelerometerData.add(it.toAccelerometerData())
                                }
                                gyroscopeDataLocal.forEach {
                                    this.gyroscopeData.add(it.toGyroscopeData())
                                }
                            }

                            bundleMap.put(
                                id,
                                BundleLocal(
                                    appLifeCycleEvents,
                                    touchTypeData,
                                    keyPressEvents,
                                    inputChangeEvents,
                                    focusChangeData,
                                    viewPortScrollData,
                                    frameRateEvents,
                                    accelerometerDataLocal,
                                    gyroscopeDataLocal
                                )
                            )

                            val bundleAsBase64: String =
                                if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
                                    Base64.getEncoder().encodeToString(bundle.toByteArray())
                                } else {
                                    android.util.Base64.encodeToString(
                                        bundle.toByteArray(),
                                        android.util.Base64.DEFAULT
                                    )
                                }
                            Log.d(webSocket.TAG, ": $bundle")

                            Log.d(webSocket.TAG, ": $bundleAsBase64")

                            try {
                                webSocket.webSocketService.send(
                                    JSONObject()
                                        .put("action", "bundle")
                                        .put("subSessionId", subSessionId)
                                        .put("payload", bundleAsBase64)
                                        .toString()
                                )
                            } catch (throwable: Throwable) {
                                handleException(throwable)
                            }

                            delay(PERIODIC_BUNDLE_UPLOAD_TIME_MS)
                            bundleIndex++
                        }
                    }
                }
            }

            job6 = coroutineScopeIO.launch {
                webSocket.webSocketMessage.collect {
                    try {
                        val jsonObject = JSONObject(it)
                        if (!jsonObject.has("error")) {
                            val bundleId = jsonObject.getString("bundleId")
                            removeInfoFromBundleMap(bundleId)
                        }
                    } catch (throwable: Throwable) {
                        handleException(throwable)
                    }
                }
            }
        }
    }

    private suspend fun removeInfoFromBundleMap(bundleId: String) {
        val bundle = bundleMap.get(bundleId)
        if (bundle != null) {
            bbRepository.deleteAppLifecycleEventList(bundle.appLifeCycleEvents)
            bbRepository.deleteFocusChangeDataList(bundle.focusChangeData)
            bbRepository.deleteFrameRateEventList(bundle.frameRateEvents)
            bbRepository.deleteViewPortScrollDataList(bundle.viewPortScrollData)
            bbRepository.deleteInputChangeEventList(bundle.inputChangeEvents)
            bbRepository.deleteKeyPressEventList(bundle.keyPressEvents)
            bbRepository.deleteTouchTypeDataList(bundle.touchTypeData)
            bbRepository.deleteAccelerometerDataList(bundle.accelerometerDataLocal)
            bbRepository.deleteGyroscopeDataList(bundle.gyroscopeDataLocal)
            /*bundleList.removeIf {
                it.id == bundleId
            }*/
        }
        bundleMap.remove(bundleId)
    }

    @Keep
    fun stopSubSession(skipDelay: Boolean = false, lambda: (() -> Unit)? = null) {
        coroutineScopeMain.launch {
            if (!skipDelay) {
                delay(PERIODIC_BUNDLE_UPLOAD_TIME_MS)
            }
            cancelAllJobs()
            foregroundStateManager.windowCallback?.isSessionActive = false
            foregroundStateManager.windowCallback?.cancelCapturingData()
            coroutineScopeIO.launch {
                deleteAllBundleData()
                if (lambda != null) {
                    lambda()
                }
            }
        }
    }

    private suspend fun deleteAllBundleData() {
        bbRepository.deleteAllFrameRateEvents()
        bbRepository.deleteAllInputChangeEvents()
        bbRepository.deleteAllFocusChangeData()
        bbRepository.deleteAllAppLifecycleEvents()
        bbRepository.deleteAllViewPortScrollData()
        bbRepository.deleteAllTouchTypeData()
        bbRepository.deleteAllKeyPressEvents()
        bbRepository.deleteAllAccelerometerData()
        bbRepository.deleteAllGyroscopeData()
    }

    private val bundleMap = mutableMapOf<String, BundleLocal>()

    private fun cancelAllJobs() {
        job1?.cancel()
        job2?.cancel()
        job3?.cancel()
        job4?.cancel()
        job5?.cancel()
        job6?.cancel()
    }

    private fun handleException(throwable: Throwable) {
        try {
            if (sendExceptionToSentry) {
                Sentry.captureException(throwable)
            }
        } catch (throwable: Throwable) {
            Log.e("Sentry", "Failed to report a crash.", throwable)
        }
        /*if (BuildConfig.DEBUG) {
            Log.e("Sentry", "Something went wrong.", throwable)
            throw RuntimeException(throwable)
        }*/
    }

    @Keep
    companion object {
        internal lateinit var INSTANCE: UserBehaviorCapture
        private const val PERIODIC_BUNDLE_UPLOAD_TIME_MS = 1000L

        @JvmStatic
        fun init(
            application: Application,
            environment: Environment,
            credId: String,
            sessionId: String,
            sendExceptionToSentry: Boolean,
            isDev: Boolean
        ): UserBehaviorCapture {
            return UserBehaviorCapture(application, environment, credId, sessionId, sendExceptionToSentry, isDev)
        }
    }
}
