package com.bureau.behavioralbiometrics

import android.app.Activity
import android.util.Log
import android.view.ActionMode
import android.view.KeyEvent
import android.view.Menu
import android.view.MenuItem
import android.view.MotionEvent
import android.view.SearchEvent
import android.view.View
import android.view.Window
import android.view.WindowManager
import android.view.accessibility.AccessibilityEvent
import androidx.core.view.GestureDetectorCompat
import com.bureau.behavioralbiometrics.data.BBRepository
import com.bureau.behavioralbiometrics.di.COROUTINE_SCOPE_IO
import com.bureau.behavioralbiometrics.focuschangedata.FocusChangeDataDetector
import com.bureau.behavioralbiometrics.keypressTypedata.KeyPressTypeDetector
import com.bureau.behavioralbiometrics.touchtypedata.TouchTypeDetector
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Job
import kotlinx.coroutines.launch
import javax.inject.Inject
import javax.inject.Named

class WindowCallback(activity: Activity) : Window.Callback {
    @Inject
    lateinit var touchTypeDetector: TouchTypeDetector

    @Inject
    lateinit var gestureDetector: GestureDetectorCompat

    @Inject
    lateinit var bbRepository: BBRepository

    @Inject
    @Named(COROUTINE_SCOPE_IO)
    lateinit var coroutineScopeIO: CoroutineScope

    @Inject
    lateinit var callback: Window.Callback

    @Inject
    lateinit var keyPressTypeDetector: KeyPressTypeDetector

    @Inject
    lateinit var focusChangeDataDetector: FocusChangeDataDetector

    private var job1: Job? = null
    private var job2: Job? = null
    private var job3: Job? = null
    private var job4: Job? = null
    private var job5: Job? = null

    var isCapturing = false
    var isSessionActive = false

    init {
        UserBehaviorCapture.INSTANCE.bbComponent.bbActivityComponent().create(activity).inject(this)
    }

    fun captureData() {
        cancelCapturingData()
        isCapturing = true
        job1 = coroutineScopeIO.launch {
            touchTypeDetector.touchTypeDataLocalFlow.collect {
                bbRepository.insertTouchTypeData(it)
                Log.d("TouchTypeData", ": $it")
            }
        }

        job2 = coroutineScopeIO.launch {
            touchTypeDetector.viewPortScrollDataLocalFlow.collect {
                bbRepository.insertViewPortScrollData(it)
                Log.d("ViewPortScrollData", ": $it")
            }
        }

        job3 = coroutineScopeIO.launch {
            keyPressTypeDetector.keyPressEventLocalFlow.collect {
                bbRepository.insertKeyPressEvent(it)
                Log.d("KeyPressEvent", ": $it")
            }
        }

        job4 = coroutineScopeIO.launch {
            keyPressTypeDetector.inputChangeEventLocalFlow.collect {
                bbRepository.insertInputChangeEvent(it)
                Log.d("InputChangeEvent", ": $it")
            }
        }

        job5 = coroutineScopeIO.launch {
            focusChangeDataDetector.focusChangeDataLocalFlow.collect {
                bbRepository.insertFocusChangeData(it)
                Log.d("FocusChangeData", ": $it")
            }
        }
    }

    fun cancelCapturingData() {
        job1?.cancel()
        job2?.cancel()
        job3?.cancel()
        job4?.cancel()
        job5?.cancel()
        isCapturing = false
    }

    override fun dispatchGenericMotionEvent(motionEvent: MotionEvent): Boolean {
        return callback.dispatchGenericMotionEvent(motionEvent)
    }

    override fun dispatchKeyEvent(var1: KeyEvent): Boolean {
        if (isSessionActive && isCapturing) {
            keyPressTypeDetector.onKeyEvent(var1)
        }
        return callback.dispatchKeyEvent(var1)
    }

    override fun dispatchKeyShortcutEvent(var1: KeyEvent): Boolean {
        return callback.dispatchKeyShortcutEvent(var1)
    }

    override fun dispatchPopulateAccessibilityEvent(var1: AccessibilityEvent): Boolean {
        return callback.dispatchPopulateAccessibilityEvent(var1)
    }

    override fun dispatchTouchEvent(var1: MotionEvent): Boolean {
        if (isSessionActive && isCapturing) {
            gestureDetector.onTouchEvent(var1)
            touchTypeDetector.onTouchEvent(var1, gestureDetector)
        }
        return callback.dispatchTouchEvent(var1)
    }

    override fun dispatchTrackballEvent(var1: MotionEvent): Boolean {
        return callback.dispatchTrackballEvent(var1)
    }

    override fun onActionModeFinished(var1: ActionMode) {
        callback.onActionModeFinished(var1)
    }

    override fun onActionModeStarted(var1: ActionMode) {
        callback.onActionModeStarted(var1)
    }

    override fun onAttachedToWindow() {
        callback.onAttachedToWindow()
    }

    override fun onContentChanged() {
        callback.onContentChanged()
    }

    override fun onCreatePanelMenu(var1: Int, var2: Menu): Boolean {
        return callback.onCreatePanelMenu(var1, var2)
    }

    override fun onCreatePanelView(var1: Int): View? {
        return callback.onCreatePanelView(var1)
    }

    override fun onDetachedFromWindow() {
        callback.onDetachedFromWindow()
    }

    override fun onMenuItemSelected(var1: Int, var2: MenuItem): Boolean {
        return callback.onMenuItemSelected(var1, var2)
    }

    override fun onMenuOpened(var1: Int, var2: Menu): Boolean {
        return callback.onMenuOpened(var1, var2)
    }

    override fun onPanelClosed(var1: Int, var2: Menu) {
        callback.onPanelClosed(var1, var2)
    }

    override fun onPreparePanel(var1: Int, var2: View?, var3: Menu): Boolean {
        return callback.onPreparePanel(var1, var2, var3)
    }

    override fun onSearchRequested(): Boolean {
        return callback.onSearchRequested()
    }

    override fun onSearchRequested(var1: SearchEvent): Boolean {
        return false
    }

    override fun onWindowAttributesChanged(var1: WindowManager.LayoutParams) {
        callback.onWindowAttributesChanged(var1)
    }

    override fun onWindowFocusChanged(var1: Boolean) {
        callback.onWindowFocusChanged(var1)
    }

    override fun onWindowStartingActionMode(var1: ActionMode.Callback): ActionMode? {
        return callback.onWindowStartingActionMode(var1)
    }

    override fun onWindowStartingActionMode(var1: ActionMode.Callback, var2: Int): ActionMode? {
        return null
    }
}
