package com.bureau.behavioralbiometrics.applifecycleevents

import android.app.Activity
import android.app.ActivityManager
import android.app.Application
import android.content.Context
import android.content.Context.SENSOR_SERVICE
import android.hardware.Sensor
import android.hardware.SensorEvent
import android.hardware.SensorEventListener
import android.hardware.SensorManager
import android.os.Bundle
import android.os.Process
import android.text.TextUtils
import android.view.Choreographer
import com.bureau.behavioralbiometrics.WindowCallback
import com.bureau.behavioralbiometrics.applifecycleevents.models.AppLifecycleStatus
import com.bureau.behavioralbiometrics.di.BBScope
import com.bureau.behavioralbiometrics.di.COROUTINE_SCOPE_IO
import com.bureau.behavioralbiometrics.frameratecollection.FPSFrameCallback
import com.bureau.behavioralbiometrics.sensordata.models.AccelerometerDataLocal
import com.bureau.behavioralbiometrics.sensordata.models.GyroscopeDataLocal
import com.bureau.behavioralbiometrics.utils.TimeUtils
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.delay
import kotlinx.coroutines.flow.MutableSharedFlow
import kotlinx.coroutines.flow.SharedFlow
import kotlinx.coroutines.launch
import javax.inject.Inject
import javax.inject.Named

@BBScope
class ForegroundStateManager @Inject constructor(
    application: Application,
    @Named(COROUTINE_SCOPE_IO) private val coroutineScopeIO: CoroutineScope,
    val fpsFrameCallback: FPSFrameCallback
) {

    private var activeActivities = 0
    private var activeActivitiesToBeDestroyed = 0

    var windowCallback: WindowCallback? = null
    private val sensorManager: SensorManager = application.getSystemService(SENSOR_SERVICE) as SensorManager
    private val accelerometer = sensorManager.getDefaultSensor(Sensor.TYPE_ACCELEROMETER)
    private val gyroscope = sensorManager.getDefaultSensor(Sensor.TYPE_GYROSCOPE)

    init {
        if (checkIsMainProcess(application)) {
            application.registerActivityLifecycleCallbacks(object :
                    Application.ActivityLifecycleCallbacks, SensorEventListener {
                    override fun onActivityCreated(activity: Activity, savedInstanceState: Bundle?) {
                        windowCallback = WindowCallback(activity)
                        activity.window.callback = windowCallback
                    }

                    override fun onActivityStarted(activity: Activity) {
                        if (windowCallback?.isCapturing == false && windowCallback?.isSessionActive == true) {
                            windowCallback?.captureData()
                        }
                        activeActivities++
                        activeActivitiesToBeDestroyed++
                        emitForegroundState()
                        enableOrDisableFrameRateCalculation()
                    }

                    override fun onActivityResumed(activity: Activity) {
                        sensorManager.registerListener(this, accelerometer, SensorManager.SENSOR_DELAY_NORMAL)
                        sensorManager.registerListener(this, gyroscope, SensorManager.SENSOR_DELAY_NORMAL)
                    }

                    override fun onActivityPaused(activity: Activity) {
                        sensorManager.unregisterListener(this)
                    }

                    override fun onActivityStopped(activity: Activity) {
                        if (activeActivities >= 0) {
                            activeActivities--
                        }
                        if (activeActivities == 0) {
                            windowCallback?.cancelCapturingData()
                        }
                        emitForegroundState()
                        enableOrDisableFrameRateCalculation()
                    }

                    override fun onActivitySaveInstanceState(activity: Activity, outState: Bundle) {}

                    override fun onActivityDestroyed(activity: Activity) {
                        if (activeActivities >= 0) {
                            activeActivitiesToBeDestroyed--
                        }
                        if (activeActivities == 0) {
                            windowCallback = null
                        }
                        emitForegroundState()
                        enableOrDisableFrameRateCalculation()
                    }

                    override fun onSensorChanged(sensorEvent: SensorEvent?) {
                        val x = sensorEvent?.values?.get(0)?.toDouble()
                        val y = sensorEvent?.values?.get(1)?.toDouble()
                        val z = sensorEvent?.values?.get(2)?.toDouble()
                        val sensorType = sensorEvent?.sensor?.type
                        if (x != null && y != null && z != null) {
                            if (sensorType == Sensor.TYPE_ACCELEROMETER) {
                                coroutineScopeIO.launch {
                                    _accelerometerDataFlow.emit(
                                        AccelerometerDataLocal(
                                            timeStamp = TimeUtils.getTimeSinceLastRebootMillis(),
                                            xAcc = x,
                                            yAcc = y,
                                            zAcc = z
                                        )
                                    )
                                }
                            } else if (sensorType == Sensor.TYPE_GYROSCOPE) {
                                coroutineScopeIO.launch {
                                    _gyroscopeDataFlow.emit(
                                        GyroscopeDataLocal(
                                            timeStamp = TimeUtils.getTimeSinceLastRebootMillis(),
                                            xRot = x,
                                            yRot = y,
                                            zRot = z
                                        )
                                    )
                                }
                            }
                        }
                    }

                    override fun onAccuracyChanged(p0: Sensor?, p1: Int) {
                    }
                })
        }
    }

    private val _foregroundStateFlow = MutableSharedFlow<AppLifecycleStatus>()
    val foregroundStateFlow: SharedFlow<AppLifecycleStatus>
        get() = _foregroundStateFlow

    private val _accelerometerDataFlow = MutableSharedFlow<AccelerometerDataLocal>()
    val accelerometerDataFlow: SharedFlow<AccelerometerDataLocal>
        get() = _accelerometerDataFlow

    private val _gyroscopeDataFlow = MutableSharedFlow<GyroscopeDataLocal>()
    val gyroscopeDataFlow: SharedFlow<GyroscopeDataLocal>
        get() = _gyroscopeDataFlow

    private fun emitForegroundState() {
        coroutineScopeIO.launch {
            delay(200)
            _foregroundStateFlow.emit(getAppLifecycleEvent())
        }
    }

    private fun checkIsMainProcess(app: Application): Boolean {
        val activityManager = app.getSystemService(Context.ACTIVITY_SERVICE) as ActivityManager
        val runningAppProcessInfoList = activityManager.runningAppProcesses ?: return false
        var currProcessName: String? = null
        val currPid = Process.myPid()
        // find the process name
        for (processInfo in runningAppProcessInfoList) {
            if (null != processInfo && processInfo.pid == currPid) {
                currProcessName = processInfo.processName
            }
        }

        // is current process the main process
        return TextUtils.equals(currProcessName, app.packageName)
    }

    private fun getAppLifecycleEvent(): AppLifecycleStatus {
        return if (activeActivitiesToBeDestroyed == 0) AppLifecycleStatus.DESTROY else if (activeActivities > 0) AppLifecycleStatus.FOREGROUND else AppLifecycleStatus.BACKGROUND
    }

    private fun enableOrDisableFrameRateCalculation() {
        if (getAppLifecycleEvent() == AppLifecycleStatus.FOREGROUND) {
            fpsFrameCallback.setEnabled(true)
            Choreographer.getInstance().postFrameCallback(fpsFrameCallback)
        } else {
            fpsFrameCallback.setEnabled(false)
        }
    }
}
