package com.bureau.behavioralbiometrics.keypressTypedata

import android.app.Activity
import android.os.SystemClock
import android.util.Log
import android.view.ActionMode
import android.view.KeyEvent
import android.view.Menu
import android.view.MenuItem
import android.view.View
import android.view.ViewGroup
import android.widget.EditText
import androidx.core.view.children
import androidx.core.widget.doAfterTextChanged
import com.bureau.base.extensions.getFromClipBoard
import com.bureau.behavioralbiometrics.di.BBActivityScope
import com.bureau.behavioralbiometrics.di.COROUTINE_SCOPE_IO
import com.bureau.behavioralbiometrics.di.COROUTINE_SCOPE_MAIN
import com.bureau.behavioralbiometrics.keypressTypedata.models.InputChangeAction
import com.bureau.behavioralbiometrics.keypressTypedata.models.InputChangeEventLocal
import com.bureau.behavioralbiometrics.keypressTypedata.models.KeyPressEventLocal
import com.bureau.behavioralbiometrics.keypressTypedata.models.KeyPressType
import com.bureau.behavioralbiometrics.models.TargetElement
import com.bureau.behavioralbiometrics.utils.StringUtils.masked
import com.bureau.behavioralbiometrics.utils.ViewUtils.getId
import com.bureau.behavioralbiometrics.utils.ViewUtils.getType
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.delay
import kotlinx.coroutines.flow.MutableSharedFlow
import kotlinx.coroutines.flow.SharedFlow
import kotlinx.coroutines.isActive
import kotlinx.coroutines.launch
import javax.inject.Inject
import javax.inject.Named

@BBActivityScope
class KeyPressTypeDetector @Inject constructor(
    activity: Activity,
    @Named(COROUTINE_SCOPE_MAIN) coroutineScopeMain: CoroutineScope,
    @Named(COROUTINE_SCOPE_IO) private val coroutineScopeIo: CoroutineScope
) {
    val TAG = "KeyPressDetector"

    private val _keyPressEventLocalFlow: MutableSharedFlow<KeyPressEventLocal> = MutableSharedFlow()
    val keyPressEventLocalFlow: SharedFlow<KeyPressEventLocal>
        get() = _keyPressEventLocalFlow

    private val _inputChangeEventLocalFlow: MutableSharedFlow<InputChangeEventLocal> = MutableSharedFlow()
    val inputChangeEventLocalFlow: SharedFlow<InputChangeEventLocal>
        get() = _inputChangeEventLocalFlow

    data class EditTextInfo(val view: View, var prevText: String)
    private val editTextList = mutableListOf<EditTextInfo>()

    init {
        coroutineScopeMain.launch {
            while (isActive) {
                val viewGroup = activity.window.decorView
                    .findViewById<ViewGroup>(android.R.id.content)
                    .getChildAt(0)
                if (viewGroup != null) {
                    val viewList = mutableListOf<View>()
                    getChildren(viewGroup, viewList)
                    viewList.forEach { view ->
                        if (view is EditText && editTextList.indexOfFirst { it.view == view } == -1) {
                            editTextList.add(EditTextInfo(view, ""))
                            view.setOnKeyListener { editText, _, keyEvent ->
                                coroutineScopeIo.launch {
                                    _keyPressEventLocalFlow.emit(
                                        KeyPressEventLocal(
                                            timeStamp = SystemClock.elapsedRealtime(),
                                            keyPressType = when (keyEvent.action) {
                                                KeyEvent.ACTION_DOWN -> KeyPressType.KEY_DOWN
                                                KeyEvent.ACTION_UP -> KeyPressType.KEY_UP
                                                else -> KeyPressType.UNKNOWN
                                            },
                                            maskedKey = keyEvent.keyCode,
                                            targetElement = TargetElement(
                                                getId(editText),
                                                getType(editText)
                                            )
                                        )
                                    )
                                }
                                false
                            }

                            view.customSelectionActionModeCallback = object : ActionMode.Callback {
                                override fun onCreateActionMode(p0: ActionMode, p1: Menu): Boolean {
                                    return true
                                }

                                override fun onPrepareActionMode(
                                    p0: ActionMode,
                                    p1: Menu
                                ): Boolean {
                                    return true
                                }

                                override fun onActionItemClicked(
                                    p0: ActionMode,
                                    p1: MenuItem
                                ): Boolean {
                                    var inputChangeAction: InputChangeAction? = null
                                    when (p1.itemId) {
                                        android.R.id.copy -> {
                                            inputChangeAction = InputChangeAction.COPY
                                        }

                                        android.R.id.paste -> {
                                            inputChangeAction = InputChangeAction.PASTE
                                        }

                                        android.R.id.cut -> {
                                            inputChangeAction = InputChangeAction.CUT
                                        }

                                        android.R.id.autofill -> {
                                            inputChangeAction = InputChangeAction.AUTOFILL
                                        }

                                        android.R.id.replaceText -> {
                                            inputChangeAction = InputChangeAction.REPLACEMENT
                                        }
                                    }
                                    inputChangeAction?.let {
                                        coroutineScopeIo.launch {
                                            _inputChangeEventLocalFlow.emit(
                                                InputChangeEventLocal(
                                                    timeStamp = SystemClock.elapsedRealtime(),
                                                    targetElement = TargetElement(
                                                        getId(view),
                                                        getType(view)
                                                    ),
                                                    focus = view.isFocused,
                                                    maskedText = view.text.toString().masked(),
                                                    inputChangeAction = it
                                                )
                                            )
                                        }
                                    }
                                    return true
                                }

                                override fun onDestroyActionMode(p0: ActionMode) {
                                }
                            }
                            view.doAfterTextChanged { editable ->
                                val index = editTextList.indexOfFirst {
                                    it.view == view
                                }
                                val prevText = editTextList.get(index).prevText
                                val currText = editable.toString()
                                val inputChangeAction: InputChangeAction =
                                    if ((currText.indexOf(prevText) == 0 && currText.length - prevText.length == 1) || (
                                        prevText.indexOf(
                                                currText
                                            ) == 0 && prevText.length - currText.length == 1
                                        )
                                    ) {
                                        InputChangeAction.UNKNOWN
                                    } else if (view.context.getFromClipBoard() == currText) {
                                        InputChangeAction.PASTE
                                    } else if (currText.length < prevText.length) {
                                        InputChangeAction.CUT
                                    } else {
                                        InputChangeAction.AUTOFILL
                                    }
                                coroutineScopeIo.launch {
                                    Log.d("InputChangeEvent2", ": $editable")
                                    _inputChangeEventLocalFlow.emit(
                                        InputChangeEventLocal(
                                            timeStamp = SystemClock.elapsedRealtime(),
                                            targetElement = TargetElement(
                                                getId(view),
                                                getType(view)
                                            ),
                                            focus = view.isFocused,
                                            maskedText = editable.toString().masked(),
                                            inputChangeAction = inputChangeAction
                                        )
                                    )
                                    Log.d("InputChangeEvent2", ": $editable")
                                }
                                editTextList.get(index).prevText = currText
                            }
                        }
                    }
                }

                delay(PERIODIC_VIEW_LIST_UPDATE_TIME)
            }
        }
    }
    fun onKeyEvent(keyEvent: KeyEvent) {
        coroutineScopeIo.launch {
            _keyPressEventLocalFlow.emit(
                KeyPressEventLocal(
                    timeStamp = SystemClock.elapsedRealtime(),
                    keyPressType = when (keyEvent.action) {
                        KeyEvent.ACTION_DOWN -> KeyPressType.KEY_DOWN
                        KeyEvent.ACTION_UP -> KeyPressType.KEY_UP
                        else -> KeyPressType.UNKNOWN
                    },
                    maskedKey = keyEvent.keyCode,
                    targetElement = TargetElement("", "")
                )
            )
        }
    }

    companion object {
        private const val PERIODIC_VIEW_LIST_UPDATE_TIME = 300L
        private fun getChildren(view: View, viewList: MutableList<View>) {
            if (view is ViewGroup) {
                viewList.add(view)
                view.children.forEach {
                    getChildren(it, viewList)
                }
            } else {
                viewList.add(view)
            }
        }
    }
}
