package com.bureau.behavioralbiometrics.touchtypedata

import android.app.Activity
import android.graphics.Rect
import android.os.SystemClock
import android.util.Log
import android.view.GestureDetector.SimpleOnGestureListener
import android.view.MotionEvent
import android.view.View
import android.view.ViewGroup
import android.widget.AbsListView
import android.widget.GridView
import android.widget.ListView
import android.widget.ScrollView
import androidx.core.view.GestureDetectorCompat
import androidx.core.view.ScrollingView
import androidx.core.widget.NestedScrollView
import androidx.recyclerview.widget.RecyclerView
import com.bureau.behavioralbiometrics.di.BBActivityScope
import com.bureau.behavioralbiometrics.di.COROUTINE_SCOPE_IO
import com.bureau.behavioralbiometrics.models.TargetElement
import com.bureau.behavioralbiometrics.touchtypedata.models.ClosedRange
import com.bureau.behavioralbiometrics.touchtypedata.models.Gesture
import com.bureau.behavioralbiometrics.touchtypedata.models.Offset2D
import com.bureau.behavioralbiometrics.touchtypedata.models.Position2D
import com.bureau.behavioralbiometrics.touchtypedata.models.TouchToolType
import com.bureau.behavioralbiometrics.touchtypedata.models.TouchTypeDataLocal
import com.bureau.behavioralbiometrics.touchtypedata.models.ViewPortScrollDataLocal
import com.bureau.behavioralbiometrics.utils.TimeUtils
import com.bureau.behavioralbiometrics.utils.ViewUtils
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.flow.MutableSharedFlow
import kotlinx.coroutines.flow.SharedFlow
import kotlinx.coroutines.launch
import javax.inject.Inject
import javax.inject.Named

@BBActivityScope
class TouchTypeDetector @Inject constructor(private val activity: Activity, @Named(COROUTINE_SCOPE_IO) private val coroutineScopeIO: CoroutineScope) : SimpleOnGestureListener() {

    private val _touchTypeDataLocalFlow: MutableSharedFlow<TouchTypeDataLocal> = MutableSharedFlow()
    val touchTypeDataLocalFlow: SharedFlow<TouchTypeDataLocal>
        get() = _touchTypeDataLocalFlow

    private val _viewPortScrollDataLocalFlow = MutableSharedFlow<ViewPortScrollDataLocal>()

    val viewPortScrollDataLocalFlow: SharedFlow<ViewPortScrollDataLocal>
        get() = _viewPortScrollDataLocalFlow

    private val scrollableViews = mutableListOf<View>()

    override fun onDoubleTap(event: MotionEvent): Boolean {
        sendTouchTypeEvent(event, event, 2, Gesture.DOUBLE_TAP)
        return super.onDoubleTap(event)
    }

    override fun onFling(
        startevent: MotionEvent,
        finishevent: MotionEvent,
        velocityX: Float,
        velocityY: Float
    ): Boolean {
        sendTouchTypeEvent(startevent, finishevent, gesture = Gesture.FLING)
        return false
    }

    override fun onScroll(
        startevent: MotionEvent,
        finishevent: MotionEvent,
        distanceX: Float,
        distanceY: Float
    ): Boolean {
        val view = getViewWhereMotionEventOccurred(startevent)
        sendTouchTypeEvent(startevent, finishevent, gesture = Gesture.SCROLL)
        Log.d("ScrollChangedListener", "onScroll: ${view?.let { ViewUtils.getType(it) }}")

        if (view != null && !scrollableViews.contains(view)) {
            scrollableViews.add(view)
            when (view) {
                is AbsListView -> {
                    val scrollDistanceCalculator = ListScrollDistanceCalculator()
                    view.setOnScrollListener(scrollDistanceCalculator)
                    scrollDistanceCalculator.setScrollDistanceListener(object :
                            ListScrollDistanceCalculator.ScrollDistanceListener {
                            override fun onScrollDistanceChanged(delta: Int, total: Int) {
                                emitViewPortScrollData(view, 0, delta.toLong())
                            }
                        })
                }

                is RecyclerView -> {
                    view.addOnScrollListener(object : RecyclerView.OnScrollListener() {
                        override fun onScrolled(recyclerView: RecyclerView, dx: Int, dy: Int) {
                            super.onScrolled(recyclerView, dx, dy)
                            emitViewPortScrollData(view, dx.toLong(), dy.toLong())
                        }
                    })
                }

                is NestedScrollView -> {
                    view.setOnScrollChangeListener(
                        NestedScrollView.OnScrollChangeListener { v, scrollX, scrollY, oldScrollX, oldScrollY ->
                            emitViewPortScrollData(view, (scrollX - oldScrollX).toLong(), (scrollY - oldScrollY).toLong())
                        }
                    )
                }
            }
        }
        return super.onScroll(startevent, finishevent, distanceX, distanceY)
    }

    private fun emitViewPortScrollData(view: View, deltaX: Long, deltaY: Long) {
        Log.d("ScrollChangedListener", ": $deltaX")
        Log.d("ScrollChangedListener", ": $deltaY")
        coroutineScopeIO.launch {
            _viewPortScrollDataLocalFlow.emit(
                ViewPortScrollDataLocal(
                    timeStamp = TimeUtils.getTimeSinceLastRebootMillis(),
                    deltaX = deltaX,
                    deltaY = deltaY,
                    targetElement = TargetElement(
                        ViewUtils.getId(view),
                        ViewUtils.getType(view)
                    )
                )
            )
        }
    }

    override fun onSingleTapConfirmed(event: MotionEvent): Boolean {
        sendTouchTypeEvent(event, event, gesture = Gesture.SINGE_TAP)
        return super.onSingleTapConfirmed(event)
    }

    override fun onLongPress(event: MotionEvent) {
        sendTouchTypeEvent(event, event, gesture = Gesture.LONG_PRESS)
        super.onLongPress(event)
    }

    private fun sendTouchTypeEvent(
        var1: MotionEvent,
        var2: MotionEvent,
        count: Int = 1,
        gesture: Gesture
    ) {
        val var3 = getViewWhereMotionEventOccurred(var1)
        for (i in 0 until var1.pointerCount) {
            val touchTypeDataLocal = TouchTypeDataLocal(
                timeStamp = SystemClock.elapsedRealtime(),
                deviceId = var1.getPointerId(i).toLong(),
                offset2D = Offset2D((var1.getX(i) - var2.getX(i)).toDouble(), (var1.getY(i) - var2.getY(i)).toDouble()),
                touchToolType = TouchToolType.getTouchToolTypeFromToolType(var1.getToolType(i)),
                position2D = Position2D(var2.getX(i).toDouble(), var2.getY(i).toDouble()),
                pressure = var2.getPressure(i).toDouble(),
                pressureRange = ClosedRange(0.0, 1.0),
                radiusMajor = var2.getTouchMajor(i).toDouble(),
                radiusMinor = var2.getToolMinor(i).toDouble(),
                orientation = var2.getOrientation(i).toDouble(),
                targetElement = if (var3 != null) {
                    TargetElement(
                        ViewUtils.getId(var3),
                        ViewUtils.getType(var3)
                    )
                } else {
                    TargetElement("", "")
                },
                gesture = gesture
            )
            coroutineScopeIO.launch {
                repeat(count) {
                    _touchTypeDataLocalFlow.emit(touchTypeDataLocal)
                }
            }
        }
    }

    private fun getViewWhereMotionEventOccurred(motionEvent: MotionEvent): View? {
        val rootView = activity.window.decorView.rootView
        return getMotionEventView(rootView, motionEvent)
    }

    private fun getMotionEventView(view: View, motionEvent: MotionEvent): View? {
        return if (view !is ViewGroup) {
            if (checkIfMotionEventIsPresentInView(view, motionEvent)) {
                view
            } else {
                null
            }
        } else {
            for (i in 0 until view.childCount) {
                val childView = view.getChildAt(i)
                if (childView is ListView || childView is ScrollView || childView is GridView || childView is RecyclerView || childView is ScrollingView) {
                    if (checkIfMotionEventIsPresentInView(childView, motionEvent)) {
                        return childView
                    }
                }
                val updatedChildView = getMotionEventView(childView, motionEvent)
                if (updatedChildView != null) {
                    return updatedChildView
                }
            }
            null
        }
    }

    private fun checkIfMotionEventIsPresentInView(view: View, motionEvent: MotionEvent): Boolean {
        val intArray = IntArray(2)
        view.getLocationOnScreen(intArray)
        val rect = Rect(
            intArray[0],
            intArray[1],
            intArray[0] + view.measuredWidth,
            intArray[1] + view.measuredHeight
        )
        if (rect.contains(motionEvent.x.toInt(), motionEvent.y.toInt())) {
            return true
        }

        return false
    }

    fun onTouchEvent(event: MotionEvent?, gestureDetectorCompat: GestureDetectorCompat): Boolean {
        return if (event != null) {
            when (event.actionMasked) {
                MotionEvent.ACTION_POINTER_DOWN -> if (event.pointerCount == 3) {
                    // touchTypListener.onThreeFingerSingleTap()
                } else if (event.pointerCount == 2) {
                    // touchTypListener.onTwoFingerSingleTap()
                }
            }
            gestureDetectorCompat.onTouchEvent(event)
        } else {
            false
        }
    }
}
