package com.bureau.behavioralbiometrics.websocket

import android.util.Log
import com.bureau.base.Environment
import com.bureau.behavioralbiometrics.di.COROUTINE_SCOPE_IO
import com.bureau.behavioralbiometrics.di.CRED_ID
import com.bureau.behavioralbiometrics.di.IS_DEV
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.flow.MutableSharedFlow
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.flow.SharedFlow
import kotlinx.coroutines.flow.StateFlow
import kotlinx.coroutines.launch
import okhttp3.OkHttpClient
import okhttp3.Request
import okhttp3.Response
import okhttp3.WebSocket
import okhttp3.WebSocketListener
import okio.ByteString
import java.nio.charset.StandardCharsets
import javax.inject.Inject
import javax.inject.Named

class WebSocket @Inject constructor(@Named(CRED_ID) credId: String, @Named(COROUTINE_SCOPE_IO) coroutineScopeIO: CoroutineScope, environment: Environment, @Named(IS_DEV) isDev: Boolean) {
    val TAG = "WEBSOCKET"
    private val request = Request.Builder()
        .url(
            if (isDev) {
                String.format(URL_DEV, credId)
            } else {
                if (environment == Environment.ENV_PRODUCTION) {
                    String.format(URL_PROD, credId)
                } else {
                    String.format(URL_STG, credId)
                }
            }
        )
        .build()

    private val _webSocketStatus = MutableStateFlow(WebSocketStatus.CLOSED)

    val webSocketStatus: StateFlow<WebSocketStatus>
        get() = _webSocketStatus

    private val _webSocketMessage = MutableSharedFlow<String>()

    val webSocketMessage: SharedFlow<String>
        get() = _webSocketMessage

    private val listener = object : WebSocketListener() {
        override fun onOpen(webSocket: WebSocket, response: Response) {
            Log.d(TAG, "onOpen: $response")
            coroutineScopeIO.launch {
                _webSocketStatus.emit(WebSocketStatus.OPEN)
            }
        }

        /** Invoked when a text (type `0x1`) message has been received. */
        override fun onMessage(webSocket: WebSocket, text: String) {
            Log.d(TAG, "onMessageT: $text")
            coroutineScopeIO.launch {
                _webSocketMessage.emit(text)
            }
        }

        /** Invoked when a binary (type `0x2`) message has been received. */
        override fun onMessage(webSocket: WebSocket, bytes: ByteString) {
            Log.d(TAG, "onMessageB: $bytes")
            coroutineScopeIO.launch {
                _webSocketMessage.emit(bytes.string(StandardCharsets.UTF_8))
            }
        }

        /**
         * Invoked when the remote peer has indicated that no more incoming messages will be transmitted.
         */
        override fun onClosing(webSocket: WebSocket, code: Int, reason: String) {
            coroutineScopeIO.launch {
                _webSocketStatus.emit(WebSocketStatus.CLOSED)
            }
        }

        /**
         * Invoked when both peers have indicated that no more messages will be transmitted and the
         * connection has been successfully released. No further calls to this listener will be made.
         */
        override fun onClosed(webSocket: WebSocket, code: Int, reason: String) {
            coroutineScopeIO.launch {
                _webSocketStatus.emit(WebSocketStatus.CLOSED)
            }
        }

        /**
         * Invoked when a web socket has been closed due to an error reading from or writing to the
         * network. Both outgoing and incoming messages may have been lost. No further calls to this
         * listener will be made.
         */
        override fun onFailure(webSocket: WebSocket, t: Throwable, response: Response?) {
            Log.d(TAG, "onFailure: $t $response")
            coroutineScopeIO.launch {
                _webSocketStatus.emit(WebSocketStatus.CLOSED)
            }
        }
    }
    val webSocketService = OkHttpClient()
        .newWebSocket(request, listener)

    enum class WebSocketStatus {
        CLOSED,
        OPEN
    }

    companion object {
        const val URL_STG = "wss://api-ws.overwatch.stg.bureau.id/?credentialId=%s"
        const val URL_PROD = "wss://api-ws.overwatch.bureau.id/?credentialId=%s"
        const val URL_DEV = "wss://api-ws.overwatch.dev.bureau.id/?credentialId=%s"
    }
}
